@@ -9,6 +9,7 @@ import threading
 import time
 import os
 import random
+import typing
 from functools import wraps, partial
 from urllib.parse import urlparse
 from enum import Enum
@@ -53,6 +54,9 @@ _KEYCODE_BACK = 4
 _MIN_API_FOR_RUNTIME_PERMISSIONS = 23
 _MIN_API_FOR_DARK_MODE = 29
 
+_REGEX_BACKUP_ALLOWED = '(pkgFlags|flags).*ALLOW_BACKUP'
+_REGEX_DEBUGGABLE = '(pkgFlags|flags).*DEBUGGABLE'
+
 # Value to be return as 'on' to the user
 _USER_PRINT_VALUE_ON = 'on'
 # Value to be return as 'partially on' to the user
@@ -900,6 +904,33 @@ def get_permission_group(args):
         return None
 
 
+# Android keeps emptying these groups so that granted permissions are denied
+# but the expectation of this tool is to do the right mapping
+def _get_hardcoded_permissions_for_group(permission_group) -> typing.List[str]:
+    if permission_group == 'android.permission-group.CONTACTS':
+        return ['android.permission.READ_CONTACTS', 'android.permission.WRITE_CONTACTS']
+    elif permission_group == 'android.permission-group.PHONE':
+        return ['android.permission.READ_PHONE_STATE', 'android.permission.READ_PHONE_NUMBERS',
+                'android.permission.CALL_PHONE', 'android.permission.ANSWER_PHONE_CALLS']
+    elif permission_group == 'android.permission-group.CALENDAR':
+        return ['android.permission.READ_CALENDAR', 'android.permission.WRITE_CALENDAR']
+    elif permission_group == 'android.permission-group.CAMERA':
+        return ['android.permission.CAMERA']
+    elif permission_group == 'android.permission-group.SENSORS':
+        return ['android.permission.BODY_SENSORS']
+    elif permission_group == 'android.permission-group.LOCATION':
+        return ['android.permission.ACCESS_FINE_LOCATION', 'android.permission.ACCESS_COARSE_LOCATION']
+    elif permission_group == 'android.permission-group.STORAGE':
+        return ['android.permission.READ_EXTERNAL_STORAGE', 'android.permission.WRITE_EXTERNAL_STORAGE']
+    elif permission_group == 'android.permission-group.MICROPHONE':
+        return ['android.permission.RECORD_AUDIO']
+    elif permission_group == 'android.permission-group.SMS':
+        return ['android.permission.READ_SMS', 'android.permission.RECEIVE_SMS', 'android.permission.SEND_SMS']
+    else:
+        print_verbose('Unexpected permission group: %s' % permission_group)
+        return []
+
+
 # Pass the full-qualified permission group name to this method.
 def get_permissions_in_permission_group(permission_group):
     # List permissions by group
@@ -927,22 +958,38 @@ def get_permissions_in_permission_group(permission_group):
             permissions = list(map(
                 lambda x: x.replace(
                     'permission:', ''), permissions))
+            permissions = list(set(permissions + _get_hardcoded_permissions_for_group(permission_group)))
             print_message(
                 'Permissions in %s group are %s' %
                 (permission_group, permissions))
             return permissions
-    return None
+    return _get_hardcoded_permissions_for_group(permission_group)
 
 
 @ensure_package_exists
 def grant_or_revoke_runtime_permissions(package_name, action_grant, permissions):
     _error_if_min_version_less_than(23)
+
+    app_info_dump = execute_adb_shell_command('dumpsys package %s' % package_name)
+    permissions_formatted_dump = _get_permissions_info_above_api_23(app_info_dump).split('\n')
+
+    action_display_name = ''
     if action_grant:
-        cmd = 'pm grant %s' % package_name
+        base_cmd = 'pm grant %s' % package_name
+        action_display_name = 'Granting'
     else:
-        cmd = 'pm revoke %s' % package_name
+        base_cmd = 'pm revoke %s' % package_name
+        action_display_name = 'Revoking'
+    num_permissions_granted = 0
     for permission in permissions:
-        execute_adb_shell_command(cmd + ' ' + permission)
+        if permission not in permissions_formatted_dump:
+            print_message('Permission %s is not requested by %s, skipping' % (permission, package_name))
+            continue
+        num_permissions_granted += 1
+        print_message('%s %s permission to %s' % (action_display_name, permission, package_name))
+        execute_adb_shell_command(base_cmd + ' ' + permission)
+    if num_permissions_granted == 0:
+        print_error_and_exit('None of these permissions were granted to %s: %s' %(package_name, permissions))
 
 
 def _get_all_packages(pm_cmd):
@@ -1048,22 +1095,39 @@ def print_list_non_system_apps():
     print('\n'.join(get_list_non_system_apps()))
 
 
-def list_debug_apps():
+def get_list_debug_apps():
+    """Return a list of installed debug applications.
+    :returns: debug_packages
+        WHERE
+        list[str] debug_packages is a strings list of debuggable packages
+    :Example:
+    >>> import adbe.adb_enhanced as adb_e
+    >>> import adbe.adb_helper as adb_h
+    >>> adb_h.set_device_id("DEVICE_ID")
+    >>> list_debug_apps = adb_e.get_list_debug_apps()
+    """
     cmd = 'pm list packages'
     packages = _get_all_packages(cmd)
+    debug_packages = []
 
     if _ASYNCIO_AVAILABLE:
         method_to_call = _is_debug_package
         params_list = packages
         result_list = asyncio_helper.execute_in_parallel(method_to_call, params_list)
-        debug_packages = []
+
         for (package_name, debuggable) in result_list:
             if debuggable:
                 debug_packages.append(package_name)
-        print('\n'.join(debug_packages))
     else:
+        debug_packages = _list_debug_apps_no_async(packages)
+    return debug_packages
+
+
+def print_list_debug_apps():
+    if not _ASYNCIO_AVAILABLE:
         print_message('Use python3 for faster execution of this call')
-        _list_debug_apps_no_async(packages)
+
+    print('\n'.join(get_list_debug_apps()))
 
 
 def _list_debug_apps_no_async(packages):
@@ -1076,23 +1140,12 @@ def _list_debug_apps_no_async(packages):
         # No faster way to do this except to check each and every package individually
         if _is_debug_package(package)[1]:
             debug_packages.append(package)
-    print('\n'.join(debug_packages))
 
-
-_REGEX_DEBUGGABLE = '(pkgFlags|flags).*DEBUGGABLE'
+    return debug_packages
 
 
 def _is_debug_package(app_name):
-    pm_cmd = 'dumpsys package %s' % app_name
-    grep_cmd = '(grep -c -E \'%s\' || true)' % _REGEX_DEBUGGABLE
-    app_info_dump = execute_adb_shell_command(pm_cmd, piped_into_cmd=grep_cmd)
-    if app_info_dump is None or app_info_dump.strip() == '0':
-        return app_name, False
-    elif app_info_dump.strip() in ('1', '2'):
-        return app_name, True
-    else:
-        print_error_and_exit('Unexpected output for %s | %s = %s' % (pm_cmd, grep_cmd, app_info_dump))
-        return None, False
+    return _package_contains_flag(app_name, _REGEX_DEBUGGABLE)
 
 
 def list_allow_backup_apps():
@@ -1126,19 +1179,23 @@ def _list_allow_backup_apps_no_async(packages):
     print('\n'.join(debug_packages))
 
 
-_REGEX_BACKUP_ALLOWED = '(pkgFlags|flags).*ALLOW_BACKUP'
-
-
 def _is_allow_backup_package(app_name):
+    return _package_contains_flag(app_name, _REGEX_BACKUP_ALLOWED)
+
+
+def _package_contains_flag(app_name, flag_regex):
     pm_cmd = 'dumpsys package %s' % app_name
-    grep_cmd = '(grep -c -E \'%s\' || true)' % _REGEX_BACKUP_ALLOWED
+    grep_cmd = '(grep -c -E \'%s\' || true)' % flag_regex
     app_info_dump = execute_adb_shell_command(pm_cmd, piped_into_cmd=grep_cmd)
     if app_info_dump is None or app_info_dump.strip() == '0':
         return app_name, False
-    elif app_info_dump.strip() in ('1', '2'):
-        return app_name, True
-    else:
-        print_error_and_exit('Unexpected output for %s | %s = %s' % (pm_cmd, grep_cmd, app_info_dump))
+    try:
+        val = int(app_info_dump.strip())
+        if val > 0:
+            return app_name, True
+        return None, False
+    except ValueError:
+        print_error_and_exit('Unexpected output for %s | %s = "%s"' % (pm_cmd, grep_cmd, app_info_dump))
         return None, False
 
 
@@ -1203,7 +1260,7 @@ def list_directory(file_path, long_format, recursive, include_hidden_files):
         cmd_prefix += ' -R'
     if include_hidden_files:
         cmd_prefix += ' -a'
-    cmd = '%s %s' % (cmd_prefix, file_path)
+    cmd = '%s \"%s\"' % (cmd_prefix, file_path)
     print_message(execute_file_related_adb_shell_command(cmd, file_path))
 
 
@@ -1296,7 +1353,7 @@ def push_file(local_file_path, remote_file_path):
     if not os.path.exists(local_file_path):
         print_error_and_exit('Local file %s does not exist' % local_file_path)
     if os.path.isdir(local_file_path):
-        print_error_and_exit('This tool does not support pushing a directory yet' % local_file_path)
+        print_error_and_exit('This tool does not support pushing a directory yet: %s' % local_file_path)
 
     # First push to tmp file in /data/local/tmp and then move that
     tmp_file = _create_tmp_file()
@@ -1489,8 +1546,7 @@ def print_app_path(app_name):
 def print_app_signature(app_name):
     apk_path = _get_apk_path(app_name)
     # Copy apk to a temp file on the disk
-    tmp_apk_file = tempfile.NamedTemporaryFile(prefix=app_name, suffix='.apk')
-    with tmp_apk_file:
+    with tempfile.NamedTemporaryFile(prefix=app_name, suffix='.apk') as tmp_apk_file:
         tmp_apk_file_name = tmp_apk_file.name
         adb_cmd = 'pull %s %s' % (apk_path, tmp_apk_file_name)
         return_code, _, stderr = execute_adb_command2(adb_cmd)
@@ -1505,13 +1561,13 @@ def print_app_signature(app_name):
 
         print_signature_cmd = 'java -jar %s verify --print-certs %s' % (apk_signer_jar_path, tmp_apk_file_name)
         print_verbose('Executing command %s' % print_signature_cmd)
-        ps1 = subprocess.Popen(print_signature_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-        for line in ps1.stdout:
-            line = line.decode('utf-8').strip()
-            print_message(line)
-        for line in ps1.stderr:
-            line = line.decode('utf-8').strip()
-            print_error(line)
+        with subprocess.Popen(print_signature_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as ps1:
+            for line in ps1.stdout:
+                line = line.decode('utf-8').strip()
+                print_message(line)
+            for line in ps1.stderr:
+                line = line.decode('utf-8').strip()
+                print_error(line)
 
 
 # Uses abe.jar taken from https://sourceforge.net/projects/adbextractor/
@@ -1559,16 +1615,16 @@ def perform_app_backup(app_name, backup_tar_file):
             print_error_and_exit('Abe.jar is missing, your adb-enhanced installation is corrupted')
         abe_cmd = 'java -jar %s unpack backup.ab %s %s' % (abe_jar_path, backup_tar_file, password)
         print_verbose('Executing command %s' % abe_cmd)
-        ps = subprocess.Popen(abe_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-        ps.communicate()
-        if ps.returncode == 0:
-            print_message('Successfully backed up data of app %s to %s' % (app_name, backup_tar_file))
-        else:
-            print_error('Failed to convert backup.ab to tar file. Please ensure that it is not password protected')
+        with subprocess.Popen(abe_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as ps:
+            ps.communicate()
+            if ps.returncode == 0:
+                print_message('Successfully backed up data of app %s to %s' % (app_name, backup_tar_file))
+            else:
+                print_error('Failed to convert backup.ab to tar file. Please ensure that it is not password protected')
     finally:
         print_verbose('Deleting backup.ab')
-        ps = subprocess.Popen('rm backup.ab', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-        ps.communicate()
+        with subprocess.Popen('rm backup.ab', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as ps2:
+            ps2.communicate()
 
 
 def perform_install(file_path):
