@@ -1,6 +1,5 @@
 import functools
 import subprocess
-import sys
 
 try:
     # This fails when the code is executed directly and not as a part of python package installation,
@@ -19,10 +18,6 @@ _IGNORED_LINES = [
 # Below version 24, if an adb shell command fails, then it still has an incorrect exit code of 0.
 _MIN_VERSION_ABOVE_WHICH_ADB_SHELL_RETURNS_CORRECT_EXIT_CODE = 24
 
-# This won't be required once I deprecate Python 2.
-if sys.version_info[0] >= 3:
-    unicode = str
-
 
 def get_adb_prefix():
     return _adb_prefix
@@ -61,9 +56,9 @@ def execute_adb_command2(adb_cmd, piped_into_cmd=None, ignore_stderr=False, devi
         final_cmd = '%s | %s' % (final_cmd, piped_into_cmd)
 
     print_verbose("Executing \"%s\"" % final_cmd)
-    ps1 = subprocess.Popen(final_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    stdout_data, stderr_data = ps1.communicate()
-    return_code = ps1.returncode
+    with subprocess.Popen(final_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as ps1:
+        stdout_data, stderr_data = ps1.communicate()
+        return_code = ps1.returncode
     try:
         stdout_data = stdout_data.decode('utf-8')
     except UnicodeDecodeError:
@@ -149,11 +144,23 @@ def execute_file_related_adb_shell_command(adb_shell_cmd, file_path, piped_into_
 
 # Gets the package name given a file path.
 # Eg. if the file is in /data/data/com.foo/.../file1 then package is com.foo
-# Limitation: Does not work with the new multi-user mode on Android.
+# Or if the file is in /data/user/0/com.foo/.../file1 then package is com.foo
 def get_package(file_path):
-    if file_path and file_path.startswith('/data/data/'):
-        run_as_package = file_path.split('/')[3]
-        return run_as_package
+    if not file_path:
+        return None
+
+    if file_path.startswith('/data/data/'):
+        items = file_path.split('/')
+        if len(items) >= 4:
+            run_as_package = items[3]
+            return run_as_package
+
+    # Handles the new multi-user mode
+    if file_path.startswith('/data/user/'):
+        items = file_path.split('/')
+        if len(items) >= 5:
+            run_as_package = items[4]
+            return run_as_package
     return None
 
 
