@@ -245,7 +245,7 @@ def main():
     elif args['apps'] and args['list'] and args['third-party']:
         adb_enhanced.print_list_non_system_apps()
     elif args['apps'] and args['list'] and args['debug']:
-        adb_enhanced.list_debug_apps()
+        adb_enhanced.print_list_debug_apps()
     elif args['apps'] and args['list'] and args['backup-enabled']:
         adb_enhanced.list_allow_backup_apps()
 
@@ -377,7 +377,7 @@ def get_generic_options_from_args(args):
 def get_version():
     dir_of_this_script = os.path.split(__file__)[0]
     version_file_path = os.path.join(dir_of_this_script, _VERSION_FILE_NAME)
-    with open(version_file_path, 'r') as file_handle:
+    with open(version_file_path, 'r', encoding='UTF-8') as file_handle:
         return file_handle.read().strip()
 
 
